#ifndef VT2_CHANNELS_ALL
#define VT2_CHANNELS_ALL	16
#endif

#define SAMPLE_MAX_LEN		64
#define ORNAMENT_MAX_LEN	256

#define FLAG_T				0x01
#define FLAG_N				0x02
#define FLAG_E				0x04
#define FLAG_T_ACC			0x08
#define FLAG_NE_ACC			0x10
#define FLAG_V_INC			0x20
#define FLAG_V_DEC			0x40

#define CENTS_RANGE			32



struct VT2SampleStruct {
	
	int flags [SAMPLE_MAX_LEN];		//flags for TNE, T/NE accumulator
	int off_t [SAMPLE_MAX_LEN];		//-4095..4095
	int off_ne[SAMPLE_MAX_LEN];		//-16..15
	int volume[SAMPLE_MAX_LEN];		//0..15
	
	int loop;
	int len;

};



struct VT2OrnamentStruct {
	
	int off[ORNAMENT_MAX_LEN];		//-96..96
	
	int loop;
	int len;
	
};



const float noteFreqs[12]={4186.01f,4434.92f,4698.63f,4978.03f,5274.04f,5587.65f,5919.91f,6271.93f,6644.88f,7040.00f,7458.62f,7902.13f};

int noteTable[12*9*CENTS_RANGE];	//extra octave in array just for calculating cents fdr the last semitone



struct VT2ChannelStruct {

	bool active;

	int frame_s;
	int frame_o;

	int cents_t;

	int pitch_t;
	int pitch_base_t;

	int pitch_n;
	int pitch_base_n;

	int pitch_e;
	int pitch_base_e;

	int volume;
	int volume_base;

	int envelope_enable;
	int envelope_type;

	int mixer;

	VT2SampleStruct *sample;
	VT2OrnamentStruct *ornament;

};



struct VT2ParserStruct {

	VT2ChannelStruct ch[VT2_CHANNELS_ALL];

};



void vt2_parser_init(VT2ParserStruct *parser,float clock)
{
	int chn,cent,note,octave,period;
	float f1,f2,div;

	//make note table

	clock/=8.0f;

	div=1024.0f;

	for(octave=0;octave<9;++octave)
	{
		for(note=0;note<12;++note)
		{
			f1=clock/(noteFreqs[note]/div);

			if(f1<0) f1=0;
			if(f1>65535) f1=65535;

			period=(int)f1;

			if(f1-period>=.5f) ++period;	//closest period

			noteTable[(octave*12+note)*CENTS_RANGE]=period;
		}

		div/=2.0f;
	}

	for(octave=0;octave<8;++octave)
	{
		for(note=0;note<12;++note)
		{
			f1=(float)noteTable[(octave*12+note+0)*CENTS_RANGE];
			f2=(float)noteTable[(octave*12+note+1)*CENTS_RANGE];

			for(cent=1;cent<CENTS_RANGE;++cent)
			{
				f1+=(f2-f1)/CENTS_RANGE;

				period=(int)f1;

				if(f1-period>=.5f) ++period;	//closest period

				noteTable[(octave*12+note)*CENTS_RANGE+cent]=period;
			}
		}
	}

	//clear all channels

	for(chn=0;chn<VT2_CHANNELS_ALL;++chn)
	{
		memset(&parser->ch[chn],0,sizeof(VT2ChannelStruct));

		parser->ch[chn].mixer=0xff;
	}
}



int hex_to_num(char c)
{
	if(c>='0'&&c<='9') return c-'0';
	if(c>='a'&&c<='f') return c-'a'+10;
	if(c>='A'&&c<='F') return c-'A'+10;
	
	return 0;
}



const char* vt2_load_sample(VT2SampleStruct *sample,const char *name)
{
	FILE *file;
	char *text;
	int ptr,row,flags,size;
	
	file=fopen(name,"rb");
	
	if(!file) return "Can't open file";
	
	fseek(file,0,SEEK_END);
	size=ftell(file);
	fseek(file,0,SEEK_SET);
	
	text=(char*)malloc(size);

	fread(text,size,1,file);
	fclose(file);
	
	if(memcmp(text,"[Sample]",8))
	{
		free(text);
		return "Not a VT2 sample file";
	}
	
	memset(sample,0,sizeof(VT2SampleStruct));

	ptr=8;
	row=0;
	
	while(ptr<size)
	{
		while(ptr<size)	//skip to new line
		{
			if(text[ptr]>=0x20) break;

			++ptr;
		}

		if(ptr>=size) break;

		sample->off_t[row]=(hex_to_num(text[ptr+5])<<8)+(hex_to_num(text[ptr+6])<<4)+hex_to_num(text[ptr+7]);
		
		sample->off_ne[row]=(hex_to_num(text[ptr+11])<<4)+hex_to_num(text[ptr+12]);
		
		sample->volume[row]=hex_to_num(text[ptr+15]);
		
		if(text[ptr+4 ]=='-') sample->off_t [row]=-sample->off_t [row];
		if(text[ptr+10]=='-') sample->off_ne[row]=-sample->off_ne[row];
		
		flags=0;
		
		if(text[ptr+0 ]=='T') flags|=FLAG_T;
		if(text[ptr+1 ]=='N') flags|=FLAG_N;
		if(text[ptr+2 ]=='E') flags|=FLAG_E;
		if(text[ptr+8 ]=='^') flags|=FLAG_T_ACC;
		if(text[ptr+13]=='^') flags|=FLAG_NE_ACC;
		if(text[ptr+16]=='-') flags|=FLAG_V_DEC;
		if(text[ptr+16]=='+') flags|=FLAG_V_INC;
		
		sample->flags[row]=flags;

		if((ptr+18)<size&&text[ptr+18]=='L') sample->loop=row;

		while(ptr<size)	//skip to end of a line
		{
			if(text[ptr]<0x20) break;

			++ptr;
		}

		++row;

		if(row>=SAMPLE_MAX_LEN) break;
	}

	sample->len=row;
	
	free(text);
	
	return NULL;
}



const char* vt2_load_ornament(VT2OrnamentStruct *ornament,const char *name)
{
	FILE *file;
	char *text;
	int num,sign,ptr,pos,size;
	
	file=fopen(name,"rb");
	
	if(!file) return "Can't open file";
	
	fseek(file,0,SEEK_END);
	size=ftell(file);
	fseek(file,0,SEEK_SET);
	
	text=(char*)malloc(size);

	fread(text,size,1,file);
	fclose(file);
	
	if(memcmp(text,"[Ornament]",10))
	{
		free(text);
		return "Not a VT2 ornament file";
	}

	memset(ornament,0,sizeof(VT2OrnamentStruct));

	ptr=10;
	pos=0;

	while(ptr<size)	//skip new line
	{
		if(text[ptr]>=0x20) break;
		
		++ptr;
	}

	while(ptr<size)
	{
		if(text[ptr]<0x20) break;	//end of ornament line
		
		if(text[ptr]=='L')
		{
			ornament->loop=pos;
			
			++ptr;
		}
		
		if(text[ptr]=='-')
		{
			sign=-1;
			
			++ptr;
		}
		else
		{
			sign=1;
		}
		
		num=0;
		
		while(ptr<size)
		{
			if(text[ptr]==','||text[ptr]<0x20)
			{
				++ptr;
				
				break;
			}
			
			num=num*10+(text[ptr++]-'0');
		}
		
		ornament->off[pos++]=num*sign;
		
		if(pos>=ORNAMENT_MAX_LEN) break;
	}
	
	ornament->len=pos;
	
	free(text);
	
	return NULL;
}



int vt2_serialize_sample(VT2SampleStruct *sample,float *dst)
{
	memcpy(dst,sample,sizeof(VT2SampleStruct));

	return (sizeof(VT2SampleStruct)+sizeof(float)-1)/sizeof(float);
}



int vt2_deserialize_sample(VT2SampleStruct *sample,float *src)
{
	memcpy(sample,src,sizeof(VT2SampleStruct));

	return (sizeof(VT2SampleStruct)+sizeof(float)-1)/sizeof(float);
}



int vt2_serialize_ornament(VT2OrnamentStruct *ornament,float *dst)
{
	memcpy(dst,ornament,sizeof(VT2OrnamentStruct));

	return (sizeof(VT2OrnamentStruct)+sizeof(float)-1)/sizeof(float);
}



int vt2_deserialize_ornament(VT2OrnamentStruct *ornament,float *src)
{
	memcpy(ornament,src,sizeof(VT2OrnamentStruct));

	return (sizeof(VT2OrnamentStruct)+sizeof(float)-1)/sizeof(float);
}



void vt2_set_pitch_t(VT2ChannelStruct *ch,float note)
{
	int cent;

	cent=(int)(note*CENTS_RANGE);

	if(cent<0) cent=0;
	if(cent>8*12*CENTS_RANGE-1) cent=8*12*CENTS_RANGE-1;

	ch->cents_t=cent;
}



void vt2_set_pitch_e(VT2ChannelStruct *ch,float note)
{
	int cent;

	cent=(int)(note*CENTS_RANGE);

	if(cent<0) cent=0;
	if(cent>8*12*CENTS_RANGE-1) cent=8*12*CENTS_RANGE-1;

	ch->pitch_base_e=noteTable[cent]>>8;

	if((noteTable[cent]&0xff)>=128) ++ch->pitch_base_e;	//closest period

	ch->pitch_e=0;
}



void vt2_set_pitch_n(VT2ChannelStruct *ch,int pitch)
{
	ch->pitch_base_n=pitch;
}



void vt2_set_volume(VT2ChannelStruct *ch,int volume)
{
	ch->volume_base=volume;
}



void vt2_play(VT2ChannelStruct *dst,VT2SampleStruct *sample,VT2OrnamentStruct *ornament,int volume,float note_t,float note_e,int type,int noise)
{
	VT2ChannelStruct ch;

	memset(&ch,0,sizeof(VT2ChannelStruct));

	ch.active=true;

	ch.sample=sample;
	ch.ornament=ornament;

	ch.mixer=0xff;
	ch.envelope_type=type;

	vt2_set_pitch_t(&ch,note_t);
	vt2_set_pitch_e(&ch,note_e);
	vt2_set_pitch_n(&ch,noise);
	vt2_set_volume(&ch,volume);

	memcpy(dst,&ch,sizeof(VT2ChannelStruct));
}



void vt2_stop(VT2ParserStruct *parser,chipStruct *chip,int chn)
{
	parser->ch[chn].active=false;

	ay_write(chip,chn,7,0xff);
	ay_write(chip,chn,8,0);
}


			  
float vt2_tone_envelope_ratio(float note_t,float offset,int type)
{
	float note_e;

	note_e=note_t+offset;

	if(type==0x0a||type==0x0e) note_e+=12.0f;	//triangle always has to be an octave higher to match saw

	if(note_e<0) note_e=0;
	if(note_e>8*12-1) note_e=8*12-1;

	return note_e;
}



void vt2_update_channel(VT2ParserStruct *parser,chipStruct *chip,int chn)
{
	int cent,flags,volume,pitch_t,pitch_n,pitch_e;
	VT2ChannelStruct *ch;

	ch=&parser->ch[chn];

	if(!ch->active) return;

	cent=ch->cents_t;

	if(ch->ornament)
	{
		cent+=ch->ornament->off[ch->frame_o]*CENTS_RANGE;

		++ch->frame_o;

		if(ch->frame_o>=ch->ornament->len) ch->frame_o=ch->ornament->loop;
	}

	if(cent<0) cent=0;
	if(cent>8*12*CENTS_RANGE-1) cent=8*12*CENTS_RANGE-1;

	ch->pitch_base_t=noteTable[cent]>>4;

	if(ch->sample)
	{
		flags=ch->sample->flags[ch->frame_s];

		if(flags&FLAG_T) ch->mixer&=~0x01; else ch->mixer|=0x01;
		if(flags&FLAG_N) ch->mixer&=~0x08; else ch->mixer|=0x08;

		if(flags&FLAG_E) ch->envelope_enable=0x10; else ch->envelope_enable=0;

		pitch_t=ch->pitch_t+ch->sample->off_t[ch->frame_s];

		if(flags&FLAG_T_ACC) ch->pitch_t=pitch_t;

		pitch_t=(pitch_t+ch->pitch_base_t)&0x0fff;

		if(!ch->envelope_enable)
		{
			pitch_n=ch->sample->off_ne[ch->frame_s];

			if(pitch_n<0) pitch_n+=0x20;

			pitch_n+=ch->pitch_n;

			if(flags&FLAG_NE_ACC) ch->pitch_n=pitch_n;

			pitch_n=(pitch_n+ch->pitch_base_n)&0x1f;
		}
		else
		{
			pitch_e=ch->pitch_e+ch->sample->off_ne[ch->frame_s];

			if(flags&FLAG_NE_ACC) ch->pitch_e=pitch_e;

			pitch_e=(pitch_e+ch->pitch_base_e)&0xffff;
		}

		ch->volume=ch->sample->volume[ch->frame_s];

		if(flags&FLAG_V_INC)
		{
			++ch->volume_base;

			if(ch->volume_base>15) ch->volume_base=15;
		}

		if(flags&FLAG_V_DEC)
		{
			--ch->volume_base;

			if(ch->volume_base<-15) ch->volume_base=-15;
		}

		volume=ch->volume-(15-ch->volume_base);

		if(volume< 0) volume= 0;
		if(volume>15) volume=15;

		++ch->frame_s;

		if(ch->frame_s>=ch->sample->len) ch->frame_s=ch->sample->loop;
	}

	ay_write(chip,chn,0,pitch_t&255);
	ay_write(chip,chn,1,pitch_t>>8);
	ay_write(chip,chn,6,pitch_n);
	ay_write(chip,chn,7,ch->mixer);
	ay_write(chip,chn,8,volume|ch->envelope_enable);
	ay_write(chip,chn,11,pitch_e&255);
	ay_write(chip,chn,12,pitch_e>>8);

	if(ch->envelope_type>-1)
	{
		ay_write(chip,chn,13,ch->envelope_type);

		ch->envelope_type=-1;
	}
}

