#include <cmath>
#include <vector>

#include <windows.h>	//needed for OpenFileDialog related stuff only

using namespace std;

#include "audioeffectx.h"



#define PLUGIN_NAME				"VT2i"
#define PLUGIN_VENDOR			"v0.1 11.04.17 Shiru"
#define PLUGIN_UID				'vt2i'

#define NUM_INPUTS				0
#define NUM_OUTPUTS				2
#define NUM_PROGRAMS			16

#define SYNTH_CHANNELS			16
#define CHIP_CHANNELS_ALL		SYNTH_CHANNELS
#define VT2_CHANNELS_ALL		SYNTH_CHANNELS

#define CHUNK_SIZE				65536

#define ENVELOPE_OFFSET_RANGE	(8*12)

#define MAX_NAME_LEN			32



#include "chip.h"
#include "parser.h"



enum {
	pIdSample=0,
	pIdOrnament,
	pIdEnvType,
	pIdEnvOffset,
	pIdNoisePitch,
	pIdChipType,
	pIdChipClock,
	pIdFrameRate,
	pIdOneShot,
	pIdDrumsMode,
	pIdPolyphony,
	pIdPortaSpeed,
	pIdVelTarget,
	pIdOutputGain,
	NUM_PARAMS
};




const char* const programDefaultNames[NUM_PROGRAMS]={
	"User 1",
	"User 2",
	"User 3",
	"User 4",
	"User 5",
	"User 6",
	"User 7",
	"User 8",
	"User 9",
	"User 10",
	"User 11",
	"User 12",
	"User 13",
	"User 14",
	"User 15",
	"User 16"
};



const int pEnvTypeList[4]={0x0c,0x0e,0x08,0x0a};



const char *pEnvTypeNames[4]={
	"Saw",
	"Triangle",
	"Inverted Saw",
	"Inverted Triangle"
};



const float pChipClockList[5]={
	1000000,	//Amstrad CPC
	1773450,	//Spectrum
	1750000,	//Pentagon
	2000000,	//Atari ST
	3500000		//2x Spectrum
};



const char *pChipClockNames[5]={
	"1.0 MHz (Amstrad CPC)",
	"1.77 MHz (Spectrum)",
	"1.75 MHz (Pentagon)",
	"2.0 MHz (Atari ST)",
	"3.5 MHz (2x Spectrum)"
};



const float pFrameRateList[]={
	48.0f,		//Non fractional BPM
	48.828f,	//Pentagon
	50.0f,		//Spectrum
	60.0f,		//Atari ST
	100.0f,		//2x Spectrum
	200.0f		//Atari ST
};



const char *pFrameRateNames[6]={
	"48.0 Hz (Non fractional BPM)",
	"48.8 Hz (Pentagon)",
	"50.0 Hz (Spectrum)",
	"60.0 Hz (Atari ST)",
	"100.0 Hz (2x Spectrum)",
	"200.0 Hz (Atari ST)"
};



const char *pDrumsModeNames[13]={
	"Disabled",
	"Base C",
	"Base C#",
	"Base D",
	"Base D#",
	"Base E",
	"Base F",
	"Base F#",
	"Base G",
	"Base G#",
	"Base A",
	"Base A#",
	"Base B"
};



const char *pVelTargetNames[3]={
	"Volume",
	"Noise",
	"Envelope"
};



enum {
	mEventTypeNote=0,
	mEventTypeProgram,
	mEventTypePitchBend,
	mEventTypeModulation,
};



struct MidiQueueStruct
{
	VstInt32 type;
	VstInt32 delta;
	VstInt32 note;
	VstInt32 velocity;
	VstInt32 program;

	float depth;
};



struct SynthChannelStruct
{
	VstInt32 note;

	float freq;
	float freq_new;
	
	float frame_acc;
	float frame_add;

	VstInt32 envelope_type;
	float envelope_offset;

	bool one_shot;
};



#define TAG(s)		(float)(((s)[0])|(((s)[1])<<7)|(((s)[2])<<14)|(((s)[3]<<21)))

#define DATA		TAG("DATA")
#define PROG		TAG("PROG")
#define ETYP		TAG("ETYP")
#define EOFF		TAG("EOFF")
#define NSPT		TAG("NSPT")
#define CHIP		TAG("CHIP")
#define CCLK		TAG("CCLK")
#define FRME		TAG("FRME")
#define ONES		TAG("ONES")
#define DRUM		TAG("DRUM")
#define POLY		TAG("POLY")
#define POSP		TAG("POSP")
#define VTGT		TAG("VTGT")
#define GAIN		TAG("GAIN")
#define SPFN		TAG("SPFN")
#define SPDR		TAG("SPDR")
#define ORFN		TAG("ORFN")
#define ORDR		TAG("ORDR")
#define SPNM		TAG("SPNM")
#define ORNM        TAG("ORNM")
#define SPDT		TAG("SPDT")
#define ORDT		TAG("ORDT")
#define DONE		TAG("DONE")



class VT2i:public AudioEffectX
{
public:

	VT2i(audioMasterCallback audioMaster);

	~VT2i();

	virtual void setParameter(VstInt32 index,float value);
	virtual float getParameter(VstInt32 index);
	virtual void getParameterLabel(VstInt32 index,char *label);
	virtual void getParameterDisplay(VstInt32 index,char *text);
	virtual void getParameterName(VstInt32 index,char *text); 

	virtual VstInt32 getProgram();
	virtual void setProgram(VstInt32 program);
	virtual void getProgramName(char* name);
	virtual void setProgramName(char* name);

	virtual bool getEffectName(char *name) { strcpy(name,PLUGIN_NAME); return true; }
	virtual bool getVendorString(char *text) { strcpy(text,PLUGIN_VENDOR); return true; }
	virtual VstInt32 getVendorVersion() { return 1000; } 

	VstInt32 canDo(char* text);
	VstInt32 getNumMidiInputChannels(void);
	VstInt32 getNumMidiOutputChannels(void);

	virtual VstInt32 getChunk(void** data,bool isPreset=false);
	virtual VstInt32 setChunk(void* data,VstInt32 byteSize,bool isPreset=false);

	VstInt32 processEvents(VstEvents* ev);
	void processReplacing(float**inputs,float **outputs,VstInt32 sampleFrames);

	static DWORD WINAPI DialogThreadProcStatic(LPVOID lpParam);

protected:

	VstInt32 Program;

	float pEnvType   [NUM_PROGRAMS];
	float pEnvOffset [NUM_PROGRAMS];
	float pNoisePitch[NUM_PROGRAMS];
	float pChipType  [NUM_PROGRAMS];
	float pChipClock [NUM_PROGRAMS];
	float pFrameRate [NUM_PROGRAMS];
	float pOneShot   [NUM_PROGRAMS];
	float pDrumsMode [NUM_PROGRAMS];
	float pPolyphony [NUM_PROGRAMS];
	float pPortaSpeed[NUM_PROGRAMS];
	float pNoteCut   [NUM_PROGRAMS];
	float pVelTarget [NUM_PROGRAMS];
	float pOutputGain[NUM_PROGRAMS];

	char ProgramName[NUM_PROGRAMS][MAX_NAME_LEN];

	VT2SampleStruct Sample[NUM_PROGRAMS];
	VT2OrnamentStruct Ornament[NUM_PROGRAMS];

	char SampleName[NUM_PROGRAMS][MAX_NAME_LEN];
	char OrnamentName[NUM_PROGRAMS][MAX_NAME_LEN];

	float FloatToMultiple(float value);

	VstInt32 SavePresetChunk(float *chunk);
	void LoadPresetChunk(float *chunk);

	VstInt32 SaveStringChunk(char *str,float *dst);
	VstInt32 LoadStringChunk(char *str,float *src);

	VstInt32 SynthAllocateVoice(VstInt32 note);
	void SynthChannelChangeNote(VstInt32 chn,VstInt32 note);

	void MidiAddNote(VstInt32 delta,VstInt32 note,VstInt32 velocity);
	void MidiAddProgramChange(VstInt32 delta,VstInt32 program);
	void MidiAddPitchBend(VstInt32 delta,float depth);
	void MidiAddModulation(VstInt32 delta,float depth);
	bool MidiIsAnyKeyDown(void);

	VstInt32 GetEnvelopeOffset(void);
	VstInt32 GetNoisePitch(void);

	vector<MidiQueueStruct> MidiQueue;

	unsigned char MidiKeyState[128];

	float MidiPitchBend;
	float MidiPitchBendRange;
	float MidiModulationDepth;
	float MidiModulationCount;

	VstInt32 MidiRPNLSB;
	VstInt32 MidiRPNMSB;
	VstInt32 MidiDataLSB;
	VstInt32 MidiDataMSB;

	VstInt32 MidiModulationMSB;

	SynthChannelStruct SynthChannel[SYNTH_CHANNELS];

	float sSlideStep;

	float Chunk[CHUNK_SIZE];

	chipStruct AY;
	VT2ParserStruct Parser;

	float PrevChipType;
	float PrevChipClock;

	float ResampleAcc;
	float ResampleAdd;

	//Windows specific code

	void DialogThreadProc(void);
	void LoadSampleOrOrnamentFromFile(bool ornament);

	HANDLE DialogThread;
	VstInt32 DialogThreadState;
	VstInt32 DialogRequest;

	char DialogSampleFileName   [MAX_PATH];
	char DialogSampleDirectory  [MAX_PATH];
	char DialogOrnamentFileName [MAX_PATH];
	char DialogOrnamentDirectory[MAX_PATH];

	void ConvertFileName(char *src,char *dst);
};